from __future__ import annotations

from typing import Any, Protocol, TypeAlias, runtime_checkable


@runtime_checkable
class VolumeContext(Protocol):
    """Minimum interface for volume weak-form evaluation."""

    test: Any
    trial: Any
    w: Any


@runtime_checkable
class SurfaceContext(Protocol):
    """Minimum interface for surface weak-form evaluation."""

    v: Any
    w: Any
    detJ: Any
    normal: Any


@runtime_checkable
class FormFieldLike(Protocol):
    """Minimum interface for form fields used in weak-form evaluation."""

    N: Any
    gradN: Any
    detJ: Any
    value_dim: int


UElement: TypeAlias = Any
ParamsLike: TypeAlias = Any



from __future__ import annotations
from typing import Any, Callable, Literal, Optional, Protocol, TYPE_CHECKING, TypeAlias, TypeVar, Union
import numpy as np
import jax
import jax.numpy as jnp

from ..mesh import HexMesh, StructuredHexBox
from .dtypes import INDEX_DTYPE
from .forms import FormContext
from .space import FESpaceBase

# Shared call signatures for kernels/forms
Array = jnp.ndarray
P = TypeVar("P")

Kernel = Callable[[FormContext, P], Array]
ResidualForm = Callable[[FormContext, Array, P], Array]
ElementDofMapper = Callable[[Array], Array]

if TYPE_CHECKING:
    from ..solver import FluxSparseMatrix, SparsityPattern
else:
    FluxSparseMatrix = Any
    SparsityPattern = Any

SparseCOO: TypeAlias = tuple[Array, Array, Array, int]
LinearCOO: TypeAlias = tuple[Array, Array, int]
JacobianReturn: TypeAlias = Union[Array, FluxSparseMatrix, SparseCOO]
BilinearReturn: TypeAlias = Union[Array, FluxSparseMatrix, SparseCOO]
LinearReturn: TypeAlias = Union[Array, LinearCOO]
MassReturn: TypeAlias = Union[FluxSparseMatrix, Array]


class ElementBilinearKernel(Protocol):
    def __call__(self, ctx: FormContext) -> Array: ...


class ElementLinearKernel(Protocol):
    def __call__(self, ctx: FormContext) -> Array: ...


class ElementResidualKernel(Protocol):
    def __call__(self, ctx: FormContext, u_elem: Array) -> Array: ...


class ElementJacobianKernel(Protocol):
    def __call__(self, u_elem: Array, ctx: FormContext) -> Array: ...

ElementKernel: TypeAlias = (
    ElementBilinearKernel
    | ElementLinearKernel
    | ElementResidualKernel
    | ElementJacobianKernel
)
def _get_pattern(space: SpaceLike, *, with_idx: bool) -> SparsityPattern | None:
    if hasattr(space, "get_sparsity_pattern"):
        return space.get_sparsity_pattern(with_idx=with_idx)
    return None


def _get_elem_rows(space: SpaceLike) -> Array:
    if hasattr(space, "get_elem_rows"):
        return space.get_elem_rows()
    return space.elem_dofs.reshape(-1)


def chunk_pad_stats(n_elems: int, n_chunks: Optional[int]) -> dict[str, int | float | None]:
    """
    Compute padding overhead for chunked assembly.
    Returns dict with chunk_size, pad, n_pad, and pad_ratio.
    """
    n_elems = int(n_elems)
    if n_chunks is None or n_elems <= 0:
        return {"chunk_size": None, "pad": 0, "n_pad": n_elems, "pad_ratio": 0.0}
    n_chunks = min(int(n_chunks), n_elems)
    chunk_size = (n_elems + n_chunks - 1) // n_chunks
    pad = (-n_elems) % chunk_size
    n_pad = n_elems + pad
    pad_ratio = float(pad) / float(n_elems) if n_elems else 0.0
    return {"chunk_size": int(chunk_size), "pad": int(pad), "n_pad": int(n_pad), "pad_ratio": pad_ratio}


def _maybe_trace_pad(
    stats: dict[str, int | float | None], *, n_chunks: Optional[int], pad_trace: bool
) -> None:
    if not pad_trace or not jax.core.trace_ctx.is_top_level():
        return
    if n_chunks is None:
        return
    print(
        "[pad]",
        f"n_chunks={int(n_chunks)}",
        f"chunk_size={stats['chunk_size']}",
        f"pad={stats['pad']}",
        f"pad_ratio={stats['pad_ratio']:.4f}",
        flush=True,
    )


class BatchedAssembler:
    """
    Assemble on a fixed space with optional masking to keep shapes static.

    Use `mask` to zero padded elements while keeping input shapes fixed.
    """

    def __init__(
        self,
        space: SpaceLike,
        elem_data: Any,
        elem_dofs: Array,
        *,
        pattern: SparsityPattern | None = None,
    ) -> None:
        self.space = space
        self.elem_data = elem_data
        self.elem_dofs = elem_dofs
        self.n_elems = int(elem_dofs.shape[0])
        self.n_ldofs = int(space.n_ldofs)
        self.n_dofs = int(space.n_dofs)
        self.pattern = pattern
        self._rows = None
        self._cols = None

    @classmethod
    def from_space(
        cls,
        space: SpaceLike,
        *,
        dep: jnp.ndarray | None = None,
        pattern: SparsityPattern | None = None,
    ) -> "BatchedAssembler":
        elem_data = space.build_form_contexts(dep=dep)
        return cls(space, elem_data, space.elem_dofs, pattern=pattern)

    def make_mask(self, n_active: int) -> Array:
        n_active = max(0, min(int(n_active), self.n_elems))
        mask = np.zeros((self.n_elems,), dtype=float)
        if n_active:
            mask[:n_active] = 1.0
        return jnp.asarray(mask)

    def slice(self, n_active: int) -> "BatchedAssembler":
        n_active = max(0, min(int(n_active), self.n_elems))
        elem_data = jax.tree_util.tree_map(lambda x: x[:n_active], self.elem_data)
        elem_dofs = self.elem_dofs[:n_active]
        return BatchedAssembler(self.space, elem_data, elem_dofs, pattern=None)

    def _rows_cols(self) -> tuple[Array, Array]:
        if self.pattern is not None:
            return self.pattern.rows, self.pattern.cols
        if self._rows is None or self._cols is None:
            elem_dofs = self.elem_dofs
            n_ldofs = int(elem_dofs.shape[1])
            rows = jnp.repeat(elem_dofs, n_ldofs, axis=1).reshape(-1)
            cols = jnp.tile(elem_dofs, (1, n_ldofs)).reshape(-1)
            self._rows = rows
            self._cols = cols
        return self._rows, self._cols

    def assemble_bilinear_with_kernel(
        self, kernel: ElementBilinearKernel, *, mask: Array | None = None
    ) -> FluxSparseMatrix:
        """
        kernel(ctx) -> (n_ldofs, n_ldofs)
        """
        from ..solver import FluxSparseMatrix

        Ke = jax.vmap(kernel)(self.elem_data)
        if mask is not None:
            Ke = Ke * jnp.asarray(mask)[:, None, None]
        data = Ke.reshape(-1)
        if self.pattern is not None:
            return FluxSparseMatrix(self.pattern, data)
        rows, cols = self._rows_cols()
        return FluxSparseMatrix(rows, cols, data, n_dofs=self.n_dofs)

    def assemble_bilinear(
        self,
        form: Kernel[P],
        params: P,
        *,
        mask: Array | None = None,
        kernel: ElementBilinearKernel | None = None,
        jit: bool = True,
    ) -> FluxSparseMatrix:
        if kernel is None:
            kernel = make_element_bilinear_kernel(form, params, jit=jit)
        return self.assemble_bilinear_with_kernel(kernel, mask=mask)

    def assemble_linear_with_kernel(
        self,
        kernel: ElementLinearKernel,
        *,
        mask: Array | None = None,
        dep: jnp.ndarray | None = None,
    ) -> Array:
        """
        kernel(ctx) -> (n_ldofs,)
        """
        elem_data = self.elem_data if dep is None else self.space.build_form_contexts(dep=dep)
        Fe = jax.vmap(kernel)(elem_data)
        if mask is not None:
            Fe = Fe * jnp.asarray(mask)[:, None]
        rows = self.elem_dofs.reshape(-1)
        data = Fe.reshape(-1)
        return jax.ops.segment_sum(data, rows, self.n_dofs)

    def assemble_linear(
        self,
        form: Kernel[P],
        params: P,
        *,
        mask: Array | None = None,
        dep: jnp.ndarray | None = None,
        kernel: ElementLinearKernel | None = None,
    ) -> Array:
        if kernel is not None:
            return self.assemble_linear_with_kernel(kernel, mask=mask, dep=dep)
        elem_data = self.elem_data if dep is None else self.space.build_form_contexts(dep=dep)
        includes_measure = getattr(form, "_includes_measure", False)

        def per_element(ctx: FormContext):
            integrand = form(ctx, params)
            if includes_measure:
                return integrand.sum(axis=0)
            wJ = ctx.w * ctx.test.detJ
            return (integrand * wJ[:, None]).sum(axis=0)

        Fe = jax.vmap(per_element)(elem_data)
        if mask is not None:
            Fe = Fe * jnp.asarray(mask)[:, None]
        rows = self.elem_dofs.reshape(-1)
        data = Fe.reshape(-1)
        return jax.ops.segment_sum(data, rows, self.n_dofs)

    def assemble_mass_matrix(
        self, *, mask: Array | None = None, lumped: bool = False
    ) -> MassReturn:
        from ..solver import FluxSparseMatrix

        n_ldofs = self.n_ldofs

        def per_element(ctx: FormContext):
            N = ctx.test.N
            base = jnp.einsum("qa,qb->qab", N, N)
            if hasattr(ctx.test, "value_dim"):
                vd = int(ctx.test.value_dim)
                I = jnp.eye(vd, dtype=N.dtype)
                base = base[:, :, :, None, None] * I[None, None, None, :, :]
                base = base.reshape(base.shape[0], n_ldofs, n_ldofs)
            wJ = ctx.w * ctx.test.detJ
            return jnp.einsum("qab,q->ab", base, wJ)

        Me = jax.vmap(per_element)(self.elem_data)
        if mask is not None:
            Me = Me * jnp.asarray(mask)[:, None, None]
        data = Me.reshape(-1)
        rows, cols = self._rows_cols()

        if lumped:
            M = jnp.zeros((self.n_dofs,), dtype=data.dtype)
            M = M.at[rows].add(data)
            return M

        return FluxSparseMatrix(rows, cols, data, n_dofs=self.n_dofs)

    def assemble_residual_with_kernel(
        self, kernel: ElementResidualKernel, u: Array, *, mask: Array | None = None
    ) -> Array:
        """
        kernel(ctx, u_elem) -> (n_ldofs,)
        """
        u_elems = jnp.asarray(u)[self.elem_dofs]
        elem_res = jax.vmap(kernel)(self.elem_data, u_elems)
        if mask is not None:
            elem_res = elem_res * jnp.asarray(mask)[:, None]
        rows = self.elem_dofs.reshape(-1)
        data = elem_res.reshape(-1)
        return jax.ops.segment_sum(data, rows, self.n_dofs)

    def assemble_residual(
        self,
        res_form: ResidualForm[P],
        u: Array,
        params: P,
        *,
        mask: Array | None = None,
        kernel: ElementResidualKernel | None = None,
    ) -> Array:
        if kernel is None:
            kernel = make_element_residual_kernel(res_form, params)
        return self.assemble_residual_with_kernel(kernel, u, mask=mask)

    def assemble_jacobian_with_kernel(
        self,
        kernel: ElementJacobianKernel,
        u: Array,
        *,
        mask: Array | None = None,
        sparse: bool = True,
        return_flux_matrix: bool = False,
    ) -> JacobianReturn:
        """
        kernel(u_elem, ctx) -> (n_ldofs, n_ldofs)
        """
        from ..solver import FluxSparseMatrix  # local import to avoid circular

        u_elems = jnp.asarray(u)[self.elem_dofs]
        J_e = jax.vmap(kernel)(u_elems, self.elem_data)
        if mask is not None:
            J_e = J_e * jnp.asarray(mask)[:, None, None]
        data = J_e.reshape(-1)
        if sparse:
            if self.pattern is not None:
                if return_flux_matrix:
                    return FluxSparseMatrix(self.pattern, data)
                return self.pattern.rows, self.pattern.cols, data, self.n_dofs
            rows, cols = self._rows_cols()
            if return_flux_matrix:
                return FluxSparseMatrix(rows, cols, data, n_dofs=self.n_dofs)
            return rows, cols, data, self.n_dofs
        rows, cols = self._rows_cols()
        idx = (rows.astype(jnp.int64) * int(self.n_dofs) + cols.astype(jnp.int64)).astype(INDEX_DTYPE)
        n_entries = self.n_dofs * self.n_dofs
        sdn = jax.lax.ScatterDimensionNumbers(
            update_window_dims=(),
            inserted_window_dims=(0,),
            scatter_dims_to_operand_dims=(0,),
        )
        K_flat = jnp.zeros(n_entries, dtype=data.dtype)
        K_flat = jax.lax.scatter_add(K_flat, idx[:, None], data, sdn)
        return K_flat.reshape(self.n_dofs, self.n_dofs)

    def assemble_jacobian(
        self,
        res_form: ResidualForm[P],
        u: Array,
        params: P,
        *,
        mask: Array | None = None,
        kernel: ElementJacobianKernel | None = None,
        sparse: bool = True,
        return_flux_matrix: bool = False,
    ) -> JacobianReturn:
        if kernel is None:
            kernel = make_element_jacobian_kernel(res_form, params)
        return self.assemble_jacobian_with_kernel(
            kernel,
            u,
            mask=mask,
            sparse=sparse,
            return_flux_matrix=return_flux_matrix,
        )

class SpaceLike(FESpaceBase, Protocol):
    pass


def assemble_bilinear_dense(
    space: SpaceLike,
    kernel: Kernel[P],
    params: P,
    *,
    sparse: bool = False,
    return_flux_matrix: bool = False,
) -> BilinearReturn:
    """
    Similar to scikit-fem's asm(biform, basis).
    kernel: FormContext, params -> (n_ldofs, n_ldofs)
    """
    elem_dofs = space.elem_dofs   # (n_elems, n_ldofs)
    n_dofs = space.n_dofs
    n_ldofs = space.n_ldofs

    elem_data = space.build_form_contexts()  # Pytree with leading n_elems in each field

    # apply kernel per element
    def ke_fun(ctx: FormContext):
        return kernel(ctx, params)

    K_e_all = jax.vmap(ke_fun)(elem_data)  # (n_elems, n_ldofs, n_ldofs)

    # ---- scatter into COO format ----
    # row/col indices (n_elems, n_ldofs, n_ldofs)
    pat = _get_pattern(space, with_idx=False)
    if pat is None:
        rows = jnp.repeat(elem_dofs, n_ldofs, axis=1)        # (n_elems, n_ldofs*n_ldofs)
        cols = jnp.tile(elem_dofs, (1, n_ldofs))             # (n_elems, n_ldofs*n_ldofs)
        rows = rows.reshape(-1)
        cols = cols.reshape(-1)
    else:
        rows = pat.rows
        cols = pat.cols
    data = K_e_all.reshape(-1)

    # Flatten indices for segment_sum via (row * n_dofs + col)
    idx = rows * n_dofs + cols  # (n_entries,)

    if sparse:
        if return_flux_matrix:
            from ..solver import FluxSparseMatrix  # local import to avoid circular
            return FluxSparseMatrix(rows, cols, data, n_dofs)
        return rows, cols, data, n_dofs

    n_entries = n_dofs * n_dofs
    out = jnp.zeros((n_entries,), dtype=data.dtype)
    out = out.at[idx].add(data)
    K = out.reshape(n_dofs, n_dofs)
    return K


def assemble_bilinear_form(
    space: SpaceLike,
    form: Kernel[P],
    params: P,
    *,
    pattern: SparsityPattern | None = None,
    n_chunks: Optional[int] = None,   # None -> no chunking
    dep: jnp.ndarray | None = None,
    kernel: ElementBilinearKernel | None = None,
    jit: bool = True,
    pad_trace: bool = False,
) -> FluxSparseMatrix:
    """
    Assemble a sparse bilinear form into a FluxSparseMatrix.

    Expects form(ctx, params) -> (n_q, n_ldofs, n_ldofs).
    If kernel is provided: kernel(ctx) -> (n_ldofs, n_ldofs).
    """
    from ..solver import FluxSparseMatrix

    if pattern is None:
        if hasattr(space, "get_sparsity_pattern"):
            pat = space.get_sparsity_pattern(with_idx=True)
        else:
            pat = make_sparsity_pattern(space, with_idx=True)
    else:
        pat = pattern
    elem_data = space.build_form_contexts(dep=dep)

    includes_measure = getattr(form, "_includes_measure", False)

    def per_element(ctx):
        integrand = form(ctx, params)                      # (n_q, m, m)
        if includes_measure:
            return integrand.sum(axis=0)
        wJ = ctx.w * ctx.test.detJ                         # (n_q,)
        return (integrand * wJ[:, None, None]).sum(axis=0)  # (m, m)

    if kernel is None:
        kernel = make_element_bilinear_kernel(form, params, jit=jit)

    # --- no-chunk path (your current implementation) ---
    if n_chunks is None:
        K_e_all = jax.vmap(kernel)(elem_data)  # (n_elems, m, m)
        data = K_e_all.reshape(-1)
        return FluxSparseMatrix(pat, data)

    # --- chunked path ---
    n_elems = space.elem_dofs.shape[0]
    if n_chunks <= 0:
        raise ValueError("n_chunks must be a positive integer.")
    n_chunks = min(int(n_chunks), int(n_elems))
    chunk_size = (n_elems + n_chunks - 1) // n_chunks
    stats = chunk_pad_stats(n_elems, n_chunks)
    _maybe_trace_pad(stats, n_chunks=n_chunks, pad_trace=pad_trace)
    # Ideally get m from pat (otherwise infer from one element).
    m = getattr(pat, "n_ldofs", None)
    if m is None:
        m = kernel(jax.tree_util.tree_map(lambda x: x[0], elem_data)).shape[0]

    # Pad to fixed-size chunks for JIT stability.
    pad = (-n_elems) % chunk_size
    if pad:
        elem_data_pad = jax.tree_util.tree_map(
            lambda x: jnp.concatenate([x, jnp.repeat(x[-1:], pad, axis=0)], axis=0),
            elem_data,
        )
    else:
        elem_data_pad = elem_data

    n_pad = n_elems + pad
    n_chunks = n_pad // chunk_size

    def _slice_first_dim(x, start, size):
        start_idx = (start,) + (0,) * (x.ndim - 1)
        slice_sizes = (size,) + x.shape[1:]
        return jax.lax.dynamic_slice(x, start_idx, slice_sizes)

    def chunk_fn(i):
        start = i * chunk_size
        ctx_chunk = jax.tree_util.tree_map(
            lambda x: _slice_first_dim(x, start, chunk_size),
            elem_data_pad,
        )
        Ke = jax.vmap(kernel)(ctx_chunk)             # (chunk, m, m)
        return Ke.reshape(-1)                             # (chunk*m*m,)

    data_chunks = jax.vmap(chunk_fn)(jnp.arange(n_chunks))
    data = data_chunks.reshape(-1)[: n_elems * m * m]
    return FluxSparseMatrix(pat, data)


def assemble_mass_matrix(
    space: SpaceLike,
    *,
    lumped: bool = False,
    n_chunks: Optional[int] = None,
    pad_trace: bool = False,
) -> MassReturn:
    """
    Assemble mass matrix M_ij = ∫ N_i N_j dΩ.
    Supports scalar and vector spaces. If lumped=True, rows are summed to diagonal.
    """
    from ..solver import FluxSparseMatrix  # local import to avoid circular

    ctxs = space.build_form_contexts()
    n_ldofs = space.n_ldofs

    def per_element(ctx: FormContext):
        N = ctx.test.N  # (n_q, n_nodes)
        base = jnp.einsum("qa,qb->qab", N, N)  # (n_q, n_nodes, n_nodes)
        if hasattr(ctx.test, "value_dim"):
            vd = int(ctx.test.value_dim)
            I = jnp.eye(vd, dtype=N.dtype)
            base = base[:, :, :, None, None] * I[None, None, None, :, :]
            base = base.reshape(base.shape[0], n_ldofs, n_ldofs)
        wJ = ctx.w * ctx.test.detJ
        return jnp.einsum("qab,q->ab", base, wJ)

    if n_chunks is None:
        M_e_all = jax.vmap(per_element)(ctxs)  # (n_elems, n_ldofs, n_ldofs)
        data = M_e_all.reshape(-1)
    else:
        n_elems = space.elem_dofs.shape[0]
        if n_chunks <= 0:
            raise ValueError("n_chunks must be a positive integer.")
        n_chunks = min(int(n_chunks), int(n_elems))
        chunk_size = (n_elems + n_chunks - 1) // n_chunks
        stats = chunk_pad_stats(n_elems, n_chunks)
        _maybe_trace_pad(stats, n_chunks=n_chunks, pad_trace=pad_trace)
        pad = (-n_elems) % chunk_size
        if pad:
            ctxs_pad = jax.tree_util.tree_map(
                lambda x: jnp.concatenate([x, jnp.repeat(x[-1:], pad, axis=0)], axis=0),
                ctxs,
            )
        else:
            ctxs_pad = ctxs

        n_pad = n_elems + pad
        n_chunks = n_pad // chunk_size

        def _slice_first_dim(x, start, size):
            start_idx = (start,) + (0,) * (x.ndim - 1)
            slice_sizes = (size,) + x.shape[1:]
            return jax.lax.dynamic_slice(x, start_idx, slice_sizes)

        def chunk_fn(i):
            start = i * chunk_size
            ctx_chunk = jax.tree_util.tree_map(
                lambda x: _slice_first_dim(x, start, chunk_size),
                ctxs_pad,
            )
            Me = jax.vmap(per_element)(ctx_chunk)  # (chunk, n_ldofs, n_ldofs)
            return Me.reshape(-1)

        data_chunks = jax.vmap(chunk_fn)(jnp.arange(n_chunks))
        data = data_chunks.reshape(-1)[: n_elems * n_ldofs * n_ldofs]

    elem_dofs = space.elem_dofs
    pat = _get_pattern(space, with_idx=False)
    if pat is None:
        rows = jnp.repeat(elem_dofs, n_ldofs, axis=1).reshape(-1)
        cols = jnp.tile(elem_dofs, (1, n_ldofs)).reshape(-1)
    else:
        rows = pat.rows
        cols = pat.cols

    if lumped:
        n_dofs = space.n_dofs
        M = jnp.zeros((n_dofs,), dtype=data.dtype)
        M = M.at[rows].add(data)
        return M

    return FluxSparseMatrix(rows, cols, data, n_dofs=space.n_dofs)


def assemble_linear_form(
    space: SpaceLike,
    form: Kernel[P],
    params: P,
    *,
    kernel: ElementLinearKernel | None = None,
    sparse: bool = False,
    n_chunks: Optional[int] = None,
    dep: jnp.ndarray | None = None,
    pad_trace: bool = False,
) -> LinearReturn:
    """
    Expects form(ctx, params) -> (n_q, n_ldofs) and integrates Σ_q form * wJ for RHS.
    If kernel is provided: kernel(ctx) -> (n_ldofs,).
    """
    elem_dofs = space.elem_dofs
    n_dofs = space.n_dofs
    n_ldofs = space.n_ldofs

    elem_data = space.build_form_contexts(dep=dep)

    includes_measure = getattr(form, "_includes_measure", False)

    if kernel is None:
        def per_element(ctx: FormContext):
            integrand = form(ctx, params)  # (n_q, m)
            if includes_measure:
                return integrand.sum(axis=0)
            wJ = ctx.w * ctx.test.detJ     # (n_q,)
            return (integrand * wJ[:, None]).sum(axis=0) # (m,)
    else:
        per_element = kernel

    if n_chunks is None:
        F_e_all = jax.vmap(per_element)(elem_data)            # (n_elems, m)
        data = F_e_all.reshape(-1)
    else:
        n_elems = space.elem_dofs.shape[0]
        m = n_ldofs
        if n_chunks <= 0:
            raise ValueError("n_chunks must be a positive integer.")
        n_chunks = min(int(n_chunks), int(n_elems))
        chunk_size = (n_elems + n_chunks - 1) // n_chunks
        stats = chunk_pad_stats(n_elems, n_chunks)
        _maybe_trace_pad(stats, n_chunks=n_chunks, pad_trace=pad_trace)
        pad = (-n_elems) % chunk_size
        if pad:
            elem_data_pad = jax.tree_util.tree_map(
                lambda x: jnp.concatenate([x, jnp.repeat(x[-1:], pad, axis=0)], axis=0),
                elem_data,
            )
        else:
            elem_data_pad = elem_data

        n_pad = n_elems + pad
        n_chunks = n_pad // chunk_size

        def _slice_first_dim(x, start, size):
            start_idx = (start,) + (0,) * (x.ndim - 1)
            slice_sizes = (size,) + x.shape[1:]
            return jax.lax.dynamic_slice(x, start_idx, slice_sizes)

        def chunk_fn(i):
            start = i * chunk_size
            ctx_chunk = jax.tree_util.tree_map(
                lambda x: _slice_first_dim(x, start, chunk_size),
                elem_data_pad,
            )
            fe = jax.vmap(per_element)(ctx_chunk)  # (chunk, m)
            return fe.reshape(-1)

        data_chunks = jax.vmap(chunk_fn)(jnp.arange(n_chunks))
        data = data_chunks.reshape(-1)[: n_elems * m]

    rows = _get_elem_rows(space)

    if sparse:
        return rows, data, n_dofs

    F = jax.ops.segment_sum(data, rows, n_dofs)
    return F


def assemble_functional(space: SpaceLike, form: Kernel[P], params: P) -> jnp.ndarray:
    """
    Assemble scalar functional J = ∫ form(ctx, params) dΩ.
    Expects form(ctx, params) -> (n_q,) or (n_q, 1).
    """
    elem_data = space.build_form_contexts()

    includes_measure = getattr(form, "_includes_measure", False)

    def per_element(ctx: FormContext):
        integrand = form(ctx, params)
        if integrand.ndim == 2 and integrand.shape[1] == 1:
            integrand = integrand[:, 0]
        if includes_measure:
            return jnp.sum(integrand)
        wJ = ctx.w * ctx.test.detJ
        return jnp.sum(integrand * wJ)

    vals = jax.vmap(per_element)(elem_data)
    return jnp.sum(vals)


def assemble_jacobian_global(
    space: SpaceLike,
    res_form: ResidualForm[P],
    u: jnp.ndarray,
    params: P,
    *,
    sparse: bool = False,
    return_flux_matrix: bool = False,
) -> JacobianReturn:
    """
    Assemble Jacobian (dR/du) from element residual res_form.
    res_form(ctx, u_elem, params) -> (n_q, n_ldofs)
    """
    elem_dofs = space.elem_dofs
    n_dofs = space.n_dofs
    n_ldofs = space.n_ldofs

    elem_data = space.build_form_contexts()

    def fe_fun(u_elem, ctx: FormContext, elem_id):
        ctx_with_id = FormContext(ctx.test, ctx.trial, ctx.x_q, ctx.w, elem_id)
        integrand = res_form(ctx_with_id, u_elem, params)  # (n_q, m)
        wJ = ctx.w * ctx.test.detJ
        fe = (integrand * wJ[:, None]).sum(axis=0)  # (m,)
        return fe

    jac_fun = jax.jacrev(fe_fun, argnums=0)

    u_elems = u[elem_dofs]  # (n_elems, n_ldofs)
    elem_ids = jnp.arange(elem_dofs.shape[0], dtype=INDEX_DTYPE)
    J_e_all = jax.vmap(jac_fun)(u_elems, elem_data, elem_ids)  # (n_elems, m, m)

    pat = _get_pattern(space, with_idx=False)
    if pat is None:
        rows = jnp.repeat(elem_dofs, n_ldofs, axis=1).reshape(-1)
        cols = jnp.tile(elem_dofs, (1, n_ldofs)).reshape(-1)
    else:
        rows = pat.rows
        cols = pat.cols
    data = J_e_all.reshape(-1)

    if sparse:
        if return_flux_matrix:
            from ..solver import FluxSparseMatrix  # local import to avoid circular
            return FluxSparseMatrix(rows, cols, data, n_dofs)
        return rows, cols, data, n_dofs

    n_entries = n_dofs * n_dofs
    idx = rows * n_dofs + cols
    K_flat = jax.ops.segment_sum(data, idx, n_entries)
    return K_flat.reshape(n_dofs, n_dofs)


def assemble_jacobian_elementwise(
    space: SpaceLike,
    res_form: ResidualForm[P],
    u: jnp.ndarray,
    params: P,
    *,
    sparse: bool = False,
    return_flux_matrix: bool = False,
) -> JacobianReturn:
    """
    Assemble Jacobian with element kernels via vmap + scatter_add.
    Recompiles if n_dofs changes, but independent of element count.
    """
    from ..solver import FluxSparseMatrix  # local import to avoid circular

    elem_dofs = space.elem_dofs
    n_dofs = space.n_dofs
    n_ldofs = space.n_ldofs

    ctxs = space.build_form_contexts()

    def fe_fun(u_elem, ctx: FormContext):
        integrand = res_form(ctx, u_elem, params)
        wJ = ctx.w * ctx.test.detJ
        return (integrand * wJ[:, None]).sum(axis=0)

    jac_fun = jax.jacrev(fe_fun, argnums=0)
    u_elems = u[elem_dofs]
    J_e_all = jax.vmap(jac_fun)(u_elems, ctxs)  # (n_elems, m, m)

    pat = _get_pattern(space, with_idx=False)
    if pat is None:
        rows = jnp.repeat(elem_dofs, n_ldofs, axis=1).reshape(-1)
        cols = jnp.tile(elem_dofs, (1, n_ldofs)).reshape(-1)
    else:
        rows = pat.rows
        cols = pat.cols
    data = J_e_all.reshape(-1)

    if sparse:
        if return_flux_matrix:
            return FluxSparseMatrix(rows, cols, data, n_dofs)
        return rows, cols, data, n_dofs

    n_entries = n_dofs * n_dofs
    idx = rows * n_dofs + cols
    sdn = jax.lax.ScatterDimensionNumbers(
        update_window_dims=(),
        inserted_window_dims=(0,),
        scatter_dims_to_operand_dims=(0,),
    )
    K_flat = jnp.zeros(n_entries, dtype=data.dtype)
    K_flat = jax.lax.scatter_add(K_flat, idx[:, None], data, sdn)
    return K_flat.reshape(pat.n_dofs, pat.n_dofs)


def assemble_residual_global(
    space: SpaceLike,
    form: ResidualForm[P],
    u: jnp.ndarray,
    params: P,
    *,
    sparse: bool = False
) -> LinearReturn:
    """
    Assemble residual vector that depends on u.
    form(ctx, u_elem, params) -> (n_q, n_ldofs)
    """
    elem_dofs = space.elem_dofs
    n_dofs = space.n_dofs
    n_ldofs = space.n_ldofs

    elem_data = space.build_form_contexts()

    def per_element(ctx: FormContext, conn: jnp.ndarray, elem_id: jnp.ndarray):
        u_elem = u[conn]
        ctx_with_id = FormContext(ctx.test, ctx.trial, ctx.x_q, ctx.w, elem_id)
        integrand = form(ctx_with_id, u_elem, params)  # (n_q, m)
        wJ = ctx.w * ctx.test.detJ
        fe = (integrand * wJ[:, None]).sum(axis=0)
        return fe

    elem_ids = jnp.arange(elem_dofs.shape[0], dtype=INDEX_DTYPE)
    F_e_all = jax.vmap(per_element)(elem_data, elem_dofs, elem_ids)  # (n_elems, m)

    rows = _get_elem_rows(space)
    data = F_e_all.reshape(-1)

    if sparse:
        return rows, data, n_dofs

    F = jax.ops.segment_sum(data, rows, n_dofs)
    return F


def assemble_residual_elementwise(
    space: SpaceLike,
    res_form: ResidualForm[P],
    u: jnp.ndarray,
    params: P,
    *,
    sparse: bool = False,
) -> LinearReturn:
    """
    Assemble residual using element kernels via vmap + scatter_add.
    Recompiles if n_dofs changes, but independent of element count.
    """
    elem_dofs = space.elem_dofs
    n_dofs = space.n_dofs
    ctxs = space.build_form_contexts()

    def per_element(ctx: FormContext, u_elem: jnp.ndarray):
        integrand = res_form(ctx, u_elem, params)
        wJ = ctx.w * ctx.test.detJ
        return (integrand * wJ[:, None]).sum(axis=0)

    u_elems = u[elem_dofs]
    F_e_all = jax.vmap(per_element)(ctxs, u_elems)  # (n_elems, m)
    rows = _get_elem_rows(space)
    data = F_e_all.reshape(-1)

    if sparse:
        return rows, data, n_dofs

    sdn = jax.lax.ScatterDimensionNumbers(
        update_window_dims=(),
        inserted_window_dims=(0,),
        scatter_dims_to_operand_dims=(0,),
    )
    F = jnp.zeros(n_dofs, dtype=data.dtype)
    F = jax.lax.scatter_add(F, rows[:, None], data, sdn)
    return F


# Backward compatibility aliases (prefer assemble_*_elementwise).
assemble_jacobian_elementwise_xla = assemble_jacobian_elementwise
assemble_residual_elementwise_xla = assemble_residual_elementwise


def make_element_bilinear_kernel(
    form: Kernel[P], params: P, *, jit: bool = True
) -> ElementBilinearKernel:
    """Element kernel: (ctx) -> Ke."""

    def per_element(ctx: FormContext):
        integrand = form(ctx, params)
        if getattr(form, "_includes_measure", False):
            return integrand.sum(axis=0)
        wJ = ctx.w * ctx.test.detJ
        return (integrand * wJ[:, None, None]).sum(axis=0)

    return jax.jit(per_element) if jit else per_element


def make_element_linear_kernel(
    form: Kernel[P], params: P, *, jit: bool = True
) -> ElementLinearKernel:
    """Element kernel: (ctx) -> fe."""

    def per_element(ctx: FormContext):
        integrand = form(ctx, params)
        if getattr(form, "_includes_measure", False):
            return integrand.sum(axis=0)
        wJ = ctx.w * ctx.test.detJ
        return (integrand * wJ[:, None]).sum(axis=0)

    return jax.jit(per_element) if jit else per_element


def make_element_residual_kernel(
    res_form: ResidualForm[P], params: P
) -> ElementResidualKernel:
    """Jitted element residual kernel: (ctx, u_elem) -> fe."""

    def per_element(ctx: FormContext, u_elem: jnp.ndarray):
        integrand = res_form(ctx, u_elem, params)
        if getattr(res_form, "_includes_measure", False):
            return integrand.sum(axis=0)
        wJ = ctx.w * ctx.test.detJ
        return (integrand * wJ[:, None]).sum(axis=0)

    return jax.jit(per_element)


def make_element_jacobian_kernel(
    res_form: ResidualForm[P], params: P
) -> ElementJacobianKernel:
    """Jitted element Jacobian kernel: (ctx, u_elem) -> Ke."""

    def fe_fun(u_elem, ctx: FormContext):
        integrand = res_form(ctx, u_elem, params)
        if getattr(res_form, "_includes_measure", False):
            return integrand.sum(axis=0)
        wJ = ctx.w * ctx.test.detJ
        return (integrand * wJ[:, None]).sum(axis=0)

    return jax.jit(jax.jacrev(fe_fun, argnums=0))


def element_residual(
    res_form: ResidualForm[P], ctx: FormContext, u_elem: jnp.ndarray, params: P
) -> Any:
    """
    Element residual vector r_e(u_e) = sum_q w_q * detJ_q * res_form(ctx, u_e, params).
    Returns shape (n_ldofs,).
    """
    integrand = res_form(ctx, u_elem, params)  # (n_q, n_ldofs) or pytree
    includes_measure = getattr(res_form, "_includes_measure", False)
    if isinstance(integrand, jnp.ndarray):
        if includes_measure:
            return jnp.einsum("qa->a", integrand)
        wJ = ctx.w * ctx.test.detJ             # (n_q,)
        return jnp.einsum("qa,q->a", integrand, wJ)
    if hasattr(ctx, "fields") and ctx.fields is not None:
        def _reduce(name, val):
            if isinstance(includes_measure, dict) and includes_measure.get(name, False):
                return jnp.einsum("qa->a", val)
            wJ = ctx.w * ctx.fields[name].test.detJ
            return jnp.einsum("qa,q->a", val, wJ)

        return {name: _reduce(name, val) for name, val in integrand.items()}
    if includes_measure:
        return jax.tree_util.tree_map(lambda x: jnp.einsum("qa->a", x), integrand)
    return jax.tree_util.tree_map(lambda x: jnp.einsum("qa,q->a", x, ctx.w * ctx.test.detJ), integrand)


def element_jacobian(
    res_form: ResidualForm[P], ctx: FormContext, u_elem: jnp.ndarray, params: P
) -> Any:
    """
    Element Jacobian K_e = d r_e / d u_e (AD via jacfwd), shape (n_ldofs, n_ldofs).
    """
    def _r_elem(u_local):
        return element_residual(res_form, ctx, u_local, params)

    return jax.jacfwd(_r_elem)(u_elem)


def make_element_kernel(
    form: Kernel[P] | ResidualForm[P],
    params: P,
    *,
    kind: Literal["bilinear", "linear", "residual", "jacobian"],
    jit: bool = True,
) -> ElementKernel:
    """
    Unified entry point for element kernels.

    kind:
      - "bilinear": kernel(ctx) -> (n_ldofs, n_ldofs)
      - "linear": kernel(ctx) -> (n_ldofs,)
      - "residual": kernel(ctx, u_elem) -> (n_ldofs,)
      - "jacobian": kernel(u_elem, ctx) -> (n_ldofs, n_ldofs)
    """
    kind = kind.lower()
    if kind == "bilinear":
        return make_element_bilinear_kernel(form, params, jit=jit)
    if kind == "linear":
        def per_element(ctx: FormContext):
            integrand = form(ctx, params)
            if getattr(form, "_includes_measure", False):
                return integrand.sum(axis=0)
            wJ = ctx.w * ctx.test.detJ
            return (integrand * wJ[:, None]).sum(axis=0)

        return jax.jit(per_element) if jit else per_element
    if kind == "residual":
        return make_element_residual_kernel(form, params)
    if kind == "jacobian":
        return make_element_jacobian_kernel(form, params)
    raise ValueError(f"Unknown kernel kind: {kind}")


def make_sparsity_pattern(space: SpaceLike, *, with_idx: bool = True) -> SparsityPattern:
    """
    Build a SparsityPattern (rows/cols[/idx]) that is independent of the solution.
    NOTE: rows/cols ordering matches assemble_jacobian_values(...).reshape(-1)
    so that pattern and data are aligned 1:1. If you change the flattening/
    compression strategy, keep this ordering contract in sync.
    """
    from ..solver import SparsityPattern  # local import to avoid circular

    elem_dofs = jnp.asarray(space.elem_dofs, dtype=INDEX_DTYPE)
    n_dofs = int(space.n_dofs)
    n_ldofs = int(space.n_ldofs)

    rows = jnp.repeat(elem_dofs, n_ldofs, axis=1).reshape(-1).astype(INDEX_DTYPE)
    cols = jnp.tile(elem_dofs, (1, n_ldofs)).reshape(-1).astype(INDEX_DTYPE)

    key = rows.astype(jnp.int64) * jnp.int64(n_dofs) + cols.astype(jnp.int64)
    order = jnp.argsort(key).astype(INDEX_DTYPE)
    rows_sorted = rows[order]
    cols_sorted = cols[order]
    counts = jnp.bincount(rows_sorted, length=n_dofs).astype(INDEX_DTYPE)
    indptr_j = jnp.concatenate([jnp.array([0], dtype=INDEX_DTYPE), jnp.cumsum(counts)])
    indices_j = cols_sorted.astype(INDEX_DTYPE)
    perm = order

    if with_idx:
        idx = (rows.astype(jnp.int64) * jnp.int64(n_dofs) + cols.astype(jnp.int64)).astype(INDEX_DTYPE)
        return SparsityPattern(
            rows=rows,
            cols=cols,
            n_dofs=n_dofs,
            idx=idx,
            perm=perm,
            indptr=indptr_j,
            indices=indices_j,
        )
    return SparsityPattern(
        rows=rows,
        cols=cols,
        n_dofs=n_dofs,
        idx=None,
        perm=perm,
        indptr=indptr_j,
        indices=indices_j,
    )


def assemble_jacobian_values(
    space: SpaceLike,
    res_form: ResidualForm[P],
    u: jnp.ndarray,
    params: P,
    *,
    kernel: ElementJacobianKernel | None = None,
    n_chunks: Optional[int] = None,
    pad_trace: bool = False,
) -> Array:
    """
    Assemble only the numeric values for the Jacobian (pattern-free).
    """
    ctxs = space.build_form_contexts()
    ker = kernel if kernel is not None else make_element_jacobian_kernel(res_form, params)

    u_elems = u[space.elem_dofs]
    if n_chunks is None:
        J_e_all = jax.vmap(ker)(u_elems, ctxs)  # (n_elem, m, m)
        return J_e_all.reshape(-1)

    n_elems = int(u_elems.shape[0])
    if n_chunks <= 0:
        raise ValueError("n_chunks must be a positive integer.")
    n_chunks = min(int(n_chunks), int(n_elems))
    chunk_size = (n_elems + n_chunks - 1) // n_chunks
    stats = chunk_pad_stats(n_elems, n_chunks)
    _maybe_trace_pad(stats, n_chunks=n_chunks, pad_trace=pad_trace)
    pad = (-n_elems) % chunk_size
    if pad:
        ctxs_pad = jax.tree_util.tree_map(
            lambda x: jnp.concatenate([x, jnp.repeat(x[-1:], pad, axis=0)], axis=0),
            ctxs,
        )
        u_elems_pad = jnp.concatenate([u_elems, jnp.repeat(u_elems[-1:], pad, axis=0)], axis=0)
    else:
        ctxs_pad = ctxs
        u_elems_pad = u_elems

    n_pad = n_elems + pad
    n_chunks = n_pad // chunk_size
    m = int(space.n_ldofs)

    def _slice_first_dim(x, start, size):
        start_idx = (start,) + (0,) * (x.ndim - 1)
        slice_sizes = (size,) + x.shape[1:]
        return jax.lax.dynamic_slice(x, start_idx, slice_sizes)

    def chunk_fn(i):
        start = i * chunk_size
        ctx_chunk = jax.tree_util.tree_map(
            lambda x: _slice_first_dim(x, start, chunk_size),
            ctxs_pad,
        )
        u_chunk = _slice_first_dim(u_elems_pad, start, chunk_size)
        J_e = jax.vmap(ker)(u_chunk, ctx_chunk)
        return J_e.reshape(-1)

    data_chunks = jax.vmap(chunk_fn)(jnp.arange(n_chunks))
    return data_chunks.reshape(-1)[: n_elems * m * m]


def assemble_residual_scatter(
    space: SpaceLike,
    res_form: ResidualForm[P],
    u: jnp.ndarray,
    params: P,
    *,
    kernel: ElementResidualKernel | None = None,
    sparse: bool = False,
    n_chunks: Optional[int] = None,
    pad_trace: bool = False,
) -> LinearReturn:
    """
    Assemble residual using jitted element kernel + vmap + scatter_add.
    Avoids Python loops; good for JIT stability.

    Note: `res_form` should return the integrand only; quadrature weights and detJ
    are applied in the element kernel (make_element_residual_kernel). Do not multiply
    by w or detJ inside `res_form`.
    """
    elem_dofs = space.elem_dofs
    n_dofs = space.n_dofs
    if jax.core.trace_ctx.is_top_level():
        if np.max(elem_dofs) >= n_dofs:
            raise ValueError("elem_dofs contains index outside n_dofs")
        if np.min(elem_dofs) < 0:
            raise ValueError("elem_dofs contains negative index")
    ctxs = space.build_form_contexts()
    ker = kernel if kernel is not None else make_element_residual_kernel(res_form, params)

    u_elems = u[elem_dofs]
    if n_chunks is None:
        elem_res = jax.vmap(ker)(ctxs, u_elems)  # (n_elem, n_ldofs)
    else:
        n_elems = int(u_elems.shape[0])
        if n_chunks <= 0:
            raise ValueError("n_chunks must be a positive integer.")
        n_chunks = min(int(n_chunks), int(n_elems))
        chunk_size = (n_elems + n_chunks - 1) // n_chunks
        stats = chunk_pad_stats(n_elems, n_chunks)
        _maybe_trace_pad(stats, n_chunks=n_chunks, pad_trace=pad_trace)
        pad = (-n_elems) % chunk_size
        if pad:
            ctxs_pad = jax.tree_util.tree_map(
                lambda x: jnp.concatenate([x, jnp.repeat(x[-1:], pad, axis=0)], axis=0),
                ctxs,
            )
            u_elems_pad = jnp.concatenate([u_elems, jnp.repeat(u_elems[-1:], pad, axis=0)], axis=0)
        else:
            ctxs_pad = ctxs
            u_elems_pad = u_elems

        n_pad = n_elems + pad
        n_chunks = n_pad // chunk_size

        def _slice_first_dim(x, start, size):
            start_idx = (start,) + (0,) * (x.ndim - 1)
            slice_sizes = (size,) + x.shape[1:]
            return jax.lax.dynamic_slice(x, start_idx, slice_sizes)

        def chunk_fn(i):
            start = i * chunk_size
            ctx_chunk = jax.tree_util.tree_map(
                lambda x: _slice_first_dim(x, start, chunk_size),
                ctxs_pad,
            )
            u_chunk = _slice_first_dim(u_elems_pad, start, chunk_size)
            res_chunk = jax.vmap(ker)(ctx_chunk, u_chunk)
            return res_chunk.reshape(-1)

        data_chunks = jax.vmap(chunk_fn)(jnp.arange(n_chunks))
        elem_res = data_chunks.reshape(-1)[: n_elems * int(space.n_ldofs)].reshape(n_elems, -1)
    if jax.core.trace_ctx.is_top_level():
        if not bool(jax.block_until_ready(jnp.all(jnp.isfinite(elem_res)))):
            bad = int(jnp.count_nonzero(~jnp.isfinite(elem_res)))
            raise RuntimeError(f"[assemble_residual_scatter] elem_res nonfinite: {bad}")

    rows = _get_elem_rows(space)
    data = elem_res.reshape(-1)

    if sparse:
        return rows, data, n_dofs

    sdn = jax.lax.ScatterDimensionNumbers(
        update_window_dims=(),
        inserted_window_dims=(0,),
        scatter_dims_to_operand_dims=(0,),
    )
    F = jnp.zeros((n_dofs,), dtype=data.dtype)
    F = jax.lax.scatter_add(F, rows[:, None], data, sdn)
    return F


def assemble_jacobian_scatter(
    space: SpaceLike,
    res_form: ResidualForm[P],
    u: jnp.ndarray,
    params: P,
    *,
    kernel: ElementJacobianKernel | None = None,
    sparse: bool = False,
    return_flux_matrix: bool = False,
    pattern: SparsityPattern | None = None,
    n_chunks: Optional[int] = None,
    pad_trace: bool = False,
) -> JacobianReturn:
    """
    Assemble Jacobian using jitted element kernel + vmap + scatter_add.
    If a SparsityPattern is provided, rows/cols are reused without regeneration.
    CONTRACT: The returned `data` ordering matches `pattern.rows/cols` exactly.
    Any change to pattern generation or data flattening must preserve this.
    """
    from ..solver import FluxSparseMatrix  # local import to avoid circular

    pat = pattern if pattern is not None else make_sparsity_pattern(space, with_idx=not sparse)
    data = assemble_jacobian_values(
        space, res_form, u, params, kernel=kernel, n_chunks=n_chunks, pad_trace=pad_trace
    )

    if sparse:
        if return_flux_matrix:
            return FluxSparseMatrix(pat, data)
        return pat.rows, pat.cols, data, pat.n_dofs

    idx = pat.idx
    if idx is None:
        idx = (pat.rows.astype(jnp.int64) * int(pat.n_dofs) + pat.cols.astype(jnp.int64)).astype(INDEX_DTYPE)

    n_entries = pat.n_dofs * pat.n_dofs
    sdn = jax.lax.ScatterDimensionNumbers(
        update_window_dims=(),
        inserted_window_dims=(0,),
        scatter_dims_to_operand_dims=(0,),
    )
    K_flat = jnp.zeros(n_entries, dtype=data.dtype)
    K_flat = jax.lax.scatter_add(K_flat, idx[:, None], data, sdn)
    return K_flat.reshape(pat.n_dofs, pat.n_dofs)


# Alias scatter-based assembly as the default public API
def assemble_residual(
    space: SpaceLike,
    form: ResidualForm[P],
    u: jnp.ndarray,
    params: P,
    *,
    kernel: ElementResidualKernel | None = None,
    sparse: bool = False,
    n_chunks: Optional[int] = None,
    pad_trace: bool = False,
) -> LinearReturn:
    """
    Assemble the global residual vector (scatter-based).
    If kernel is provided: kernel(ctx, u_elem) -> (n_ldofs,).
    """
    return assemble_residual_scatter(
        space, form, u, params, kernel=kernel, sparse=sparse, n_chunks=n_chunks, pad_trace=pad_trace
    )


def assemble_jacobian(
    space: SpaceLike,
    res_form: ResidualForm[P],
    u: jnp.ndarray,
    params: P,
    *,
    kernel: ElementJacobianKernel | None = None,
    sparse: bool = True,
    return_flux_matrix: bool = False,
    pattern: SparsityPattern | None = None,
    n_chunks: Optional[int] = None,
    pad_trace: bool = False,
) -> JacobianReturn:
    """
    Assemble the global Jacobian (scatter-based).
    If kernel is provided: kernel(u_elem, ctx) -> (n_ldofs, n_ldofs).
    """
    return assemble_jacobian_scatter(
        space,
        res_form,
        u,
        params,
        kernel=kernel,
        sparse=sparse,
        return_flux_matrix=return_flux_matrix,
        pattern=pattern,
        n_chunks=n_chunks,
        pad_trace=pad_trace,
    )


def _make_unit_cube_mesh() -> HexMesh:
    """Single hex element on [0, 1]^3."""
    return StructuredHexBox(nx=1, ny=1, nz=1, lx=1.0, ly=1.0, lz=1.0).build()


def scalar_body_force_form(ctx: FormContext, load: float) -> jnp.ndarray:
    """Linear form for constant scalar body force: f * N."""
    return load * ctx.test.N  # (n_q, n_ldofs)


scalar_body_force_form._ff_kind = "linear"
scalar_body_force_form._ff_domain = "volume"


def make_scalar_body_force_form(body_force: Callable[[Array], Array]) -> Kernel[Any]:
    """
    Build a scalar linear form from a callable f(x_q) -> (n_q,).
    """
    def _form(ctx: FormContext, _params):
        f_q = body_force(ctx.x_q)
        return f_q[..., None] * ctx.test.N
    _form._ff_kind = "linear"
    _form._ff_domain = "volume"
    return _form


# Backward compatibility alias
constant_body_force_form = scalar_body_force_form


def _check_structured_box_connectivity() -> None:
    """Quick connectivity check for nx=2, ny=1, nz=1 (non-structured order)."""
    box = StructuredHexBox(nx=2, ny=1, nz=1, lx=2.0, ly=1.0, lz=1.0)
    mesh = box.build()

    assert mesh.coords.shape == (12, 3)
    assert mesh.conn.shape == (2, 8)

    expected_conn = jnp.array(
        [
            [0, 1, 4, 3, 6, 7, 10, 9],   # element at i=0
            [1, 2, 5, 4, 7, 8, 11, 10],  # element at i=1
        ],
        dtype=INDEX_DTYPE,
    )
    max_diff = int(jnp.max(jnp.abs(mesh.conn - expected_conn)))
    print("StructuredHexBox nx=2,ny=1,nz=1 conn matches expected:", max_diff == 0)
    if max_diff != 0:
        print("expected conn:\n", expected_conn)
        print("got conn:\n", mesh.conn)


if __name__ == "__main__":
    _check_structured_box_connectivity()



from __future__ import annotations

from typing import Mapping, Sequence

import numpy as np

try:
    import scipy.sparse as sp
except Exception:  # pragma: no cover
    sp = None

from .block_system import split_block_matrix
from .sparse import FluxSparseMatrix


def diag(**blocks):
    return dict(blocks)


def _infer_sizes_from_diag(diag_blocks):
    sizes = {}
    for name, blk in diag_blocks.items():
        if isinstance(blk, FluxSparseMatrix):
            sizes[name] = int(blk.n_dofs)
        elif sp is not None and sp.issparse(blk):
            shape = blk.shape
            if shape[0] != shape[1]:
                raise ValueError(f"diag block {name} must be square, got {shape}")
            sizes[name] = int(shape[0])
        else:
            arr = np.asarray(blk)
            if arr.ndim != 2 or arr.shape[0] != arr.shape[1]:
                raise ValueError(f"diag block {name} must be square, got {arr.shape}")
            sizes[name] = int(arr.shape[0])
    return sizes


def _add_blocks(a, b):
    if a is None:
        return b
    if b is None:
        return a
    if isinstance(a, FluxSparseMatrix):
        a = a.to_csr()
    if isinstance(b, FluxSparseMatrix):
        b = b.to_csr()
    if sp is not None and sp.issparse(a):
        if sp.issparse(b):
            return a + b
        return a + sp.csr_matrix(np.asarray(b))
    if sp is not None and sp.issparse(b):
        return sp.csr_matrix(np.asarray(a)) + b
    return np.asarray(a) + np.asarray(b)


def _transpose_block(block, rule: str):
    if isinstance(block, FluxSparseMatrix):
        if sp is None:
            raise ImportError("scipy is required to transpose FluxSparseMatrix blocks.")
        block = block.to_csr()
    if sp is not None and sp.issparse(block):
        out = block.T
    else:
        out = np.asarray(block).T
    if rule == "H":
        return out.conjugate()
    return out


def make(
    *,
    diag: Mapping[str, object] | Sequence[object],
    rel: Mapping[tuple[str, str], object] | None = None,
    add_contiguous: object | None = None,
    sizes: Mapping[str, int] | None = None,
    symmetric: bool = False,
    transpose_rule: str = "T",
):
    """
    Build a blocks dict from diagonal blocks, optional relations, and a full matrix.
    """
    if isinstance(diag, Mapping):
        diag_map = dict(diag)
    else:
        diag_seq = list(diag)
        if sizes is None:
            diag_map = dict(zip(range(len(diag_seq)), diag_seq))
        else:
            order = tuple(sizes.keys())
            if len(diag_seq) != len(order):
                raise ValueError("diag sequence length must match sizes")
            diag_map = dict(zip(order, diag_seq))

    if sizes is None:
        sizes = _infer_sizes_from_diag(diag_map)
    order = tuple(sizes.keys())

    if add_contiguous is None:
        blocks = {name: {} for name in order}
    else:
        blocks = split_block_matrix(add_contiguous, sizes=sizes)

    for name, blk in diag_map.items():
        if name not in sizes:
            raise KeyError(f"Unknown field '{name}' in diag")
        blocks.setdefault(name, {})
        blocks[name][name] = _add_blocks(blocks[name].get(name), blk)

    if transpose_rule not in {"T", "H", "none"}:
        raise ValueError("transpose_rule must be one of: T, H, none")

    if rel is not None:
        for (name_i, name_j), blk in rel.items():
            if name_i not in sizes or name_j not in sizes:
                raise KeyError(f"Unknown field in rel: {(name_i, name_j)}")
            blocks.setdefault(name_i, {})
            blocks[name_i][name_j] = _add_blocks(blocks[name_i].get(name_j), blk)
            if symmetric and name_i != name_j:
                if transpose_rule == "none":
                    blocks.setdefault(name_j, {})
                    blocks[name_j][name_i] = _add_blocks(blocks[name_j].get(name_i), blk)
                else:
                    blocks.setdefault(name_j, {})
                    blocks[name_j][name_i] = _add_blocks(
                        blocks[name_j].get(name_i),
                        _transpose_block(blk, transpose_rule),
                    )

    return blocks


__all__ = ["diag", "make"]



import jax

from ..core.assembly import assemble_residual, assemble_jacobian
from ..core.space import FESpace


def make_jitted_residual(space: FESpace, res_form, params, *, sparse: bool = False):
    """
    Create a jitted residual assembler: u -> R(u).
    params and space are closed over.
    """
    space_jax = space
    params_jax = params

    @jax.jit
    def residual(u):
        return assemble_residual(space_jax, res_form, u, params_jax, sparse=sparse)

    return residual


def make_jitted_jacobian(
    space: FESpace,
    res_form,
    params,
    *,
    sparse: bool = False,
    return_flux_matrix: bool = False,
):
    """
    Create a jitted Jacobian assembler: u -> J(u).
    params and space are closed over.
    """
    space_jax = space
    params_jax = params

    @jax.jit
    def jacobian(u):
        return assemble_jacobian(
            space_jax,
            res_form,
            u,
            params_jax,
            sparse=sparse,
            return_flux_matrix=return_flux_matrix,
        )

    return jacobian


__all__ = ["make_jitted_residual", "make_jitted_jacobian"]



from __future__ import annotations

import time
from dataclasses import dataclass, field
import warnings
from typing import Any, Callable, Iterable, List, Sequence

import numpy as np
import jax.numpy as jnp

from ..core.assembly import assemble_bilinear_form
from ..core.solver import spdirect_solve_cpu, spdirect_solve_gpu
from .cg import cg_solve, cg_solve_jax
from .petsc import petsc_shell_solve
from .sparse import FluxSparseMatrix
from .dirichlet import expand_dirichlet_solution
from .newton import newton_solve
from .result import SolverResult
from .history import NewtonIterRecord, LoadStepResult
from ..tools.timer import SectionTimer, NullTimer


@dataclass
class NonlinearAnalysis:
    """
    Bundle problem data needed for a Newton solve with load scaling.

    Attributes
    ----------
    space : Any
        FE space containing topology/dofs.
    residual_form : Any
        Internal residual form (e.g., neo_hookean_residual_form).
    params : Any
        Parameters forwarded to the residual form.
    base_external_vector : Any | None
        Unscaled external load vector (scaled by load factor in `external_for_load`).
    dirichlet : tuple | None
        (dofs, values) for Dirichlet boundary conditions.
    extra_terms : list[callable] | None
        Optional extra term assemblers returning (K, f[, metrics]).
    jacobian_pattern : Any | None
        Optional sparsity pattern to reuse between load steps.
    dtype : Any
        dtype for the solution vector (defaults to float64).
    """

    space: Any
    residual_form: Any
    params: Any
    base_external_vector: Any | None = None
    dirichlet: tuple | None = None
    extra_terms: list[Callable] | None = None
    jacobian_pattern: Any | None = None
    dtype: Any = jnp.float64

    def external_for_load(self, load_factor: float):
        if self.base_external_vector is None:
            return None
        return jnp.asarray(load_factor * self.base_external_vector, dtype=self.dtype)


@dataclass
class NewtonLoopConfig:
    """
    Control parameters for the Newton loop with load stepping.
    """

    tol: float = 1e-8
    atol: float = 0.0
    maxiter: int = 20
    line_search: bool = False
    max_ls: int = 10
    ls_c: float = 1e-4
    linear_solver: str = "spsolve"
    linear_maxiter: int | None = None
    linear_tol: float | None = None
    linear_preconditioner: Any | None = None
    matfree_mode: str = "linearize"
    load_sequence: Sequence[float] | None = None
    n_steps: int = 1

    def schedule(self) -> List[float]:
        """
        Return a monotonically increasing list of load factors (0->1].
        """
        if self.load_sequence is not None:
            return list(self.load_sequence)
        n = max(1, self.n_steps)
        return list(np.linspace(0.0, 1.0, n + 1, endpoint=True)[1:])


class NewtonSolveRunner:
    """
    Run one or more Newton solves across load factors.

    This orchestrates load stepping, assembles external load per step,
    and returns the full (Dirichlet-expanded) solution and per-step history.
    """

    def __init__(self, analysis: NonlinearAnalysis, config: NewtonLoopConfig):
        self.analysis = analysis
        self.config = config
        self._matfree_cache: dict[str, Any] = {}

    def run(
        self,
        u0=None,
        *,
        load_sequence: Sequence[float] | None = None,
        newton_callback: Callable | None = None,
        step_callback: Callable[[LoadStepResult], None] | None = None,
        timer: "SectionTimer | None" = None,
        report_timing: bool = True
    ):
        """
        Execute Newton solves over the configured load schedule.

        Parameters
        ----------
        u0 : array-like | None
            Initial guess (defaults to zeros).
        load_sequence : sequence | None
            Optional per-call load schedule; if omitted, uses config.schedule().
        newton_callback : callable | None
            Per-iteration callback passed to `newton_solve`.
        step_callback : callable | None
            Optional hook called after each load step with LoadStepResult.
        Returns:
            u: full solution (Dirichlet expanded), dtype per analysis.dtype
            history: list of LoadStepResult per load factor
        """
        # timer = timer or NullTimer()
        timer = timer or SectionTimer(hierarchical=True)
        with timer.section("run_total"):
            dtype = self.analysis.dtype

            with timer.section("preprocess"):

                if u0 is None:
                    u = jnp.zeros(self.analysis.space.n_dofs, dtype=dtype)
                else:
                    u = jnp.asarray(u0, dtype=dtype)

                schedule_raw = list(load_sequence) if load_sequence is not None else self.config.schedule()
                # enforce monotone increasing 0->1 schedule, warn if dropped
                schedule = []
                dropped = []
                prev = 0.0
                for lf in schedule_raw:
                    lf_clamped = float(lf)
                    if not np.isfinite(lf_clamped):
                        dropped.append(("nonfinite", lf_clamped))
                        continue
                    if lf_clamped < 0.0 or lf_clamped > 1.0:
                        dropped.append(("out_of_range", lf_clamped))
                        continue
                    if lf_clamped < prev:
                        dropped.append(("nonmonotone", lf_clamped))
                        continue
                    schedule.append(lf_clamped)
                    prev = lf_clamped
                history: List[LoadStepResult] = []
                matfree_cache = None
                if self.config.linear_preconditioner == "diag0":
                    n_free = self.analysis.space.n_dofs
                    if self.analysis.dirichlet is not None:
                        n_free -= len(self.analysis.dirichlet[0])
                    cached_free = self._matfree_cache.get("n_free_dofs")
                    if cached_free is not None and cached_free != n_free:
                        self._matfree_cache.clear()
                    self._matfree_cache["n_free_dofs"] = n_free
                    matfree_cache = self._matfree_cache
                for step_i, load_factor in enumerate(schedule, start=1):
                    with timer.section("step"):
                        external = self.analysis.external_for_load(load_factor)
                        iter_log: List[NewtonIterRecord] = []
                        ext_nnz = int(jnp.count_nonzero(external))
                        ext_inf = int(jnp.count_nonzero(external))
                        # assumes u is full dofs (xyz)
                        u_nodes0 = np.asarray(u).reshape(-1, 3)
                        max_u0 = float(np.linalg.norm(u_nodes0, axis=1).max()) if u_nodes0.size else 0.0
                        print(
                            f"[load factor step {step_i}/{len(schedule)}] lf={load_factor:.3f} "
                            f"||F||_inf={ext_inf:.3e} nnz={ext_nnz} max|u|_start={max_u0:.3e}"
                        )

                        def cb(d):
                            # d: {"iter": k, "residual_norm": ..., "rel_residual": ..., "alpha": ..., "step_norm": ...}
                            lin_iters = d.get("linear_iters")
                            lin_res = d.get("linear_residual")
                            lin_conv = d.get("linear_converged")
                            iter_log.append(
                                NewtonIterRecord(
                                    iter=int(d.get("iter", -1)),
                                    res_inf=float(d.get("res_inf", float("nan"))),
                                    res_two=float(d.get("res_two", float("nan"))),
                                    rel_res_inf=float(d.get("rel_residual", float("nan"))),
                                    alpha=float(d.get("alpha", 1.0)),
                                    step_norm=float(d.get("step_norm", float("nan"))),
                                    lin_iters=int(lin_iters) if lin_iters is not None else None,
                                    lin_converged=bool(lin_conv) if lin_conv is not None else None,
                                    lin_residual=float(lin_res) if lin_res is not None else None,
                                    nan_detected=bool(d.get("nan_detected", False)),
                                )
                            )
                            if newton_callback is not None:
                                newton_callback(d)

                    try:
                        u, info = newton_solve(
                            self.analysis.space,
                            self.analysis.residual_form,
                            u,
                            self.analysis.params,
                            tol=self.config.tol,
                            atol=self.config.atol,
                            maxiter=self.config.maxiter,
                            linear_solver=self.config.linear_solver,
                            linear_maxiter=self.config.linear_maxiter,
                            linear_tol=self.config.linear_tol,
                            linear_preconditioner=self.config.linear_preconditioner,
                            matfree_cache=matfree_cache,
                            matfree_mode=self.config.matfree_mode,
                            dirichlet=self.analysis.dirichlet,
                            line_search=self.config.line_search,
                            max_ls=self.config.max_ls,
                            ls_c=self.config.ls_c,
                            external_vector=external,
                            callback=cb,
                            jacobian_pattern=self.analysis.jacobian_pattern,
                            extra_terms=self.analysis.extra_terms,
                        )
                        exception = None
                    except Exception as e:  # pragma: no cover - defensive
                        info = SolverResult(converged=False, iters=0, stop_reason="exception", nan_detected=False)
                        exception = repr(e)

                # ===== [B] OUTER LOOP PRINT (STEP END) =====
                u_nodes1 = np.asarray(u).reshape(-1, 3)
                max_u1 = float(np.linalg.norm(u_nodes1, axis=1).max()) if u_nodes1.size else 0.0
                step_solve_time = timer._records.get("step>newton_solve", [0.0])[-1]
                print(
                    f"  -> converged={getattr(info,'converged',None)} iters={getattr(info,'iters',None)} "
                    f"time={step_solve_time:.3f}s max|u|_end={max_u1:.3e}"
                    + (f" EXC={exception}" if exception else "")
                )

                meta = {
                    "load_factor": load_factor,
                    "linear_solver": self.config.linear_solver,
                    "line_search": self.config.line_search,
                    "maxiter": self.config.maxiter,
                    "n_dofs": self.analysis.space.n_dofs,
                    "dtype": str(self.analysis.dtype),
                    "u_layout": "full",
                    "schedule": schedule,
                    "schedule_dropped": dropped,
                }

                result = LoadStepResult(
                    load_factor=load_factor,
                    info=info,
                    solve_time=step_solve_time,
                    u=u,
                    iter_history=iter_log,
                    exception=exception,
                    meta=meta,
                )
                history.append(result)
                if step_callback is not None:
                    step_callback(result)

        if report_timing:
            timer.report(sort_by="total")
        return u, history


def _condense_flux_dirichlet(K: FluxSparseMatrix, F, dirichlet):
    dir_dofs, dir_vals = dirichlet
    dir_arr = np.asarray(dir_dofs, dtype=int)
    dir_vals_arr = np.asarray(dir_vals, dtype=float)
    K_csr = K.to_csr()
    mask = np.ones(K_csr.shape[0], dtype=bool)
    mask[dir_arr] = False
    free = np.nonzero(mask)[0]
    F_full = np.asarray(F, dtype=float)
    K_fd = K_csr[free][:, dir_arr] if dir_arr.size > 0 else None
    F_free_base = F_full[free]
    offset = K_fd @ dir_vals_arr if K_fd is not None and dir_arr.size > 0 else None
    K_ff = K_csr[free][:, free]
    return K_ff, F_free_base, offset, free, dir_arr, dir_vals_arr


def solve_nonlinear(
    space,
    residual_form,
    params,
    *,
    dirichlet: tuple | None = None,
    base_external_vector=None,
    extra_terms=None,
    dtype=jnp.float64,
    maxiter: int = 20,
    tol: float = 1e-8,
    atol: float = 1e-10,
    linear_solver: str = "spsolve",
    linear_maxiter: int | None = None,
    linear_tol: float | None = None,
    linear_preconditioner=None,
    matfree_mode: str = "linearize",
    line_search: bool = False,
    max_ls: int = 10,
    ls_c: float = 1e-4,
    n_steps: int = 1,
    jacobian_pattern=None,
    u0=None,
):
    """
    Convenience wrapper: build NonlinearAnalysis and run NewtonSolveRunner.
    """
    analysis = NonlinearAnalysis(
        space=space,
        residual_form=residual_form,
        params=params,
        base_external_vector=base_external_vector,
        dirichlet=dirichlet,
        extra_terms=extra_terms,
        dtype=dtype,
        jacobian_pattern=jacobian_pattern,
    )
    cfg = NewtonLoopConfig(
        maxiter=maxiter,
        tol=tol,
        atol=atol,
        linear_solver=linear_solver,
        linear_maxiter=linear_maxiter,
        linear_tol=linear_tol,
        linear_preconditioner=linear_preconditioner,
        matfree_mode=matfree_mode,
        line_search=line_search,
        max_ls=max_ls,
        ls_c=ls_c,
        n_steps=n_steps,
    )
    runner = NewtonSolveRunner(analysis, cfg)
    u0_use = jnp.zeros(space.n_dofs, dtype=dtype) if u0 is None else u0
    u, history = runner.run(u0=u0_use)
    return u, history


@dataclass
class LinearAnalysis:
    """
    Bundle linear problem data for a single solve or a load-scaled sequence.

    The matrix is assembled once from ``bilinear_form``; the RHS is scaled
    by the load factor.
    """

    space: Any
    bilinear_form: Any
    params: Any
    base_rhs_vector: Any
    dirichlet: tuple | None = None
    pattern: Any | None = None
    dtype: Any = jnp.float64

    def assemble_matrix(self):
        return self.space.assemble_bilinear_form(
            self.bilinear_form,
            params=self.params,
            pattern=self.pattern,
        )

    def rhs_for_load(self, load_factor: float):
        return jnp.asarray(load_factor * self.base_rhs_vector, dtype=self.dtype)


@dataclass
class LinearSolveConfig:
    """
    Control parameters for the linear solve with optional load scaling.
    """

    method: str = "spsolve"  # "spsolve" | "spdirect_solve_gpu" | "cg" | "cg_custom" | "petsc_shell"
    tol: float = 1e-8
    maxiter: int | None = None
    preconditioner: Any | None = None
    ksp_type: str | None = None
    pc_type: str | None = None
    ksp_rtol: float | None = None
    ksp_atol: float | None = None
    ksp_max_it: int | None = None
    petsc_ksp_norm_type: str | None = None
    petsc_ksp_monitor_true_residual: bool = False
    petsc_ksp_converged_reason: bool = False
    petsc_ksp_monitor_short: bool = False
    petsc_shell_pmat: bool = False
    petsc_shell_pmat_mode: str = "full"
    petsc_shell_pmat_rebuild_iters: int | None = None
    petsc_shell_fallback: bool = False
    petsc_shell_fallback_ksp_types: tuple[str, ...] = ("bcgs", "gmres")
    petsc_shell_fallback_rebuild_pmat: bool = True

    @classmethod
    def from_preset(cls, name: str) -> "LinearSolveConfig":
        preset = name.lower()
        if preset == "contact":
            return cls(
                method="petsc_shell",
                ksp_type="bcgs",
                pc_type="ilu",
                petsc_shell_pmat=True,
                petsc_shell_pmat_mode="full",
                petsc_ksp_norm_type="unpreconditioned",
                petsc_ksp_monitor_true_residual=True,
                petsc_ksp_converged_reason=True,
                petsc_shell_fallback=True,
            )
        raise ValueError(f"Unknown LinearSolveConfig preset: {name}")


@dataclass
class LinearStepResult:
    """
    Result record for one linear solve step.

    Attributes
    ----------
    info : SolverResult
        Solver status and iteration metadata.
    solve_time : float
        Wall time for the solve section.
    u : Any
        Full solution vector (Dirichlet-expanded).
    """
    info: SolverResult
    solve_time: float
    u: Any


class LinearSolveRunner:
    """
    Solve linear systems for one or more load factors using a unified interface.
    """

    def __init__(self, analysis: LinearAnalysis, config: LinearSolveConfig):
        self.analysis = analysis
        self.config = config
        self._petsc_shell_pmat = None
        self._petsc_shell_last_iters = None
        self._petsc_shell_pmat_rebuilds = 0

    def run(
        self,
        *,
        step_callback: Callable[[LinearStepResult], None] | None = None,
        timer: "SectionTimer | None" = None,
        report_timing: bool = True
    ) -> tuple[np.ndarray | None, list[LinearStepResult]]:

        timer = timer or SectionTimer(hierarchical=True)
        # timer = timer or NullTimer()
        with timer.section("linear_run_total"):
            with timer.section("assemble_matrix"):
                K = self.analysis.assemble_matrix()

            with timer.section("build_rhs"):
                base_rhs = jnp.asarray(
                    self.analysis.base_rhs_vector, dtype=self.analysis.dtype
                )
                if self.analysis.dirichlet is not None:
                    K_ff, F_free_base, offset, free, dir_arr, dir_vals_arr = _condense_flux_dirichlet(
                        K, base_rhs, self.analysis.dirichlet
                    )
                    n_total = K.shape[0] if hasattr(K, "shape") else self.analysis.space.n_dofs
                else:
                    K_ff = K.to_csr()
                    F_free_base = np.asarray(base_rhs, dtype=float)
                    offset = None
                    free = None
                    dir_arr = dir_vals_arr = None
                    n_total = K_ff.shape[0]

                F_free = np.asarray(F_free_base, dtype=float)
                if offset is not None:
                    F_free = F_free - offset

            with timer.section(f"solve>{self.config.method}"):
                try:
                    if self.config.method == "spsolve":
                        u_free = spdirect_solve_cpu(K_ff, F_free)
                        lin_iters = 1
                        lin_conv = True
                        lin_res = None
                        info = SolverResult(
                            converged=True,
                            iters=lin_iters,
                            linear_iters=lin_iters,
                            linear_converged=lin_conv,
                            linear_residual=lin_res,
                            tol=self.config.tol,
                            stop_reason="converged",
                        )
                    elif self.config.method == "spdirect_solve_gpu":
                        u_free = spdirect_solve_gpu(K_ff, F_free)
                        lin_iters = 1
                        lin_conv = True
                        lin_res = None
                        info = SolverResult(
                            converged=True,
                            iters=lin_iters,
                            linear_iters=lin_iters,
                            linear_converged=lin_conv,
                            linear_residual=lin_res,
                            tol=self.config.tol,
                            stop_reason="converged",
                        )
                    elif self.config.method in ("cg", "cg_custom"):
                        coo = K_ff.tocoo()
                        A_cg = FluxSparseMatrix.from_bilinear(
                            (
                                jnp.asarray(coo.row, dtype=jnp.int32),
                                jnp.asarray(coo.col, dtype=jnp.int32),
                                jnp.asarray(coo.data),
                                K_ff.shape[0],
                            )
                        )
                        cg_solver = cg_solve_jax if self.config.method == "cg" else cg_solve
                        u_free, cg_info = cg_solver(
                            A_cg, jnp.asarray(F_free),
                            tol=self.config.tol,
                            maxiter=self.config.maxiter,
                            preconditioner=self.config.preconditioner
                        )
                        lin_iters = cg_info.get("iters")
                        lin_conv = bool(cg_info.get("converged", True))
                        lin_res = cg_info.get("residual_norm", cg_info.get("residual"))
                        info = SolverResult(
                            converged=lin_conv,
                            iters=int(lin_iters) if lin_iters is not None else 0,
                            linear_iters=int(lin_iters) if lin_iters is not None else None,
                            linear_converged=lin_conv,
                            linear_residual=float(lin_res) if lin_res is not None else None,
                            tol=self.config.tol,
                            stop_reason=("converged" if lin_conv else "linfail"),
                            nan_detected=bool(np.isnan(lin_res)) if lin_res is not None else False,
                        )
                    elif self.config.method == "petsc_shell":
                        base_ksp_type = self.config.ksp_type or "gmres"
                        pc_type = self.config.pc_type if self.config.pc_type is not None else "none"
                        ksp_rtol = self.config.ksp_rtol if self.config.ksp_rtol is not None else self.config.tol
                        ksp_atol = self.config.ksp_atol
                        ksp_max_it = self.config.ksp_max_it if self.config.ksp_max_it is not None else self.config.maxiter
                        petsc_options = {}
                        if self.config.petsc_ksp_norm_type:
                            petsc_options["fluxfem_ksp_norm_type"] = self.config.petsc_ksp_norm_type
                        if self.config.petsc_ksp_monitor_true_residual:
                            petsc_options["fluxfem_ksp_monitor_true_residual"] = ""
                        if self.config.petsc_ksp_converged_reason:
                            petsc_options["fluxfem_ksp_converged_reason"] = ""
                        if self.config.petsc_ksp_monitor_short:
                            petsc_options["fluxfem_ksp_monitor_short"] = ""
                        if not petsc_options:
                            petsc_options = None
                        use_pmat = bool(self.config.petsc_shell_pmat)
                        rebuild_thresh = self.config.petsc_shell_pmat_rebuild_iters
                        if use_pmat:
                            pmat_mode = (self.config.petsc_shell_pmat_mode or "full").lower()
                            if pmat_mode == "none":
                                use_pmat = False
                                pmat = None
                            elif pmat_mode == "full":
                                pmat = K_ff
                            else:
                                warnings.warn(
                                    f"petsc_shell_pmat_mode='{pmat_mode}' is not supported in runner; "
                                    "falling back to 'full'.",
                                    RuntimeWarning,
                                )
                                pmat = K_ff
                            if use_pmat:
                                if self._petsc_shell_pmat is None:
                                    self._petsc_shell_pmat = pmat
                                    self._petsc_shell_pmat_rebuilds += 1
                                elif rebuild_thresh is not None and self._petsc_shell_last_iters is not None:
                                    if self._petsc_shell_last_iters > rebuild_thresh:
                                        self._petsc_shell_pmat = pmat
                                        self._petsc_shell_pmat_rebuilds += 1
                                pmat = self._petsc_shell_pmat
                        if not use_pmat:
                            pmat = None

                        def _attempt_solve(ksp_type: str):
                            return petsc_shell_solve(
                                K_ff,
                                F_free,
                                preconditioner=self.config.preconditioner,
                                ksp_type=ksp_type,
                                pc_type=pc_type,
                                rtol=ksp_rtol,
                                atol=ksp_atol,
                                max_it=ksp_max_it,
                                pmat=pmat,
                                options=petsc_options,
                                return_info=True,
                            )

                        fallback_ksp = [base_ksp_type]
                        if self.config.petsc_shell_fallback:
                            for ksp in self.config.petsc_shell_fallback_ksp_types:
                                if ksp not in fallback_ksp:
                                    fallback_ksp.append(ksp)
                        fallback_attempts = []
                        petsc_info = None
                        u_free = None
                        for ksp in fallback_ksp:
                            fallback_attempts.append(ksp)
                            u_free, petsc_info = _attempt_solve(ksp)
                            lin_conv = petsc_info.get("converged")
                            reason = petsc_info.get("reason")
                            if lin_conv is None and reason is not None:
                                lin_conv = reason > 0
                            if lin_conv:
                                break
                            if self.config.petsc_shell_fallback and use_pmat and self.config.petsc_shell_fallback_rebuild_pmat:
                                self._petsc_shell_pmat = pmat
                                self._petsc_shell_pmat_rebuilds += 1
                        lin_iters = petsc_info.get("iters")
                        lin_res = petsc_info.get("residual_norm")
                        lin_solve_dt = petsc_info.get("solve_time")
                        pc_setup_dt = petsc_info.get("pc_setup_time")
                        pmat_dt = petsc_info.get("pmat_build_time")
                        lin_conv = petsc_info.get("converged")
                        if lin_conv is None and petsc_info.get("reason") is not None:
                            lin_conv = petsc_info.get("reason") > 0
                        if lin_conv is None:
                            lin_conv = True
                        self._petsc_shell_last_iters = lin_iters
                        info = SolverResult(
                            converged=bool(lin_conv),
                            iters=int(lin_iters) if lin_iters is not None else 0,
                            linear_iters=int(lin_iters) if lin_iters is not None else None,
                            linear_converged=bool(lin_conv),
                            linear_residual=float(lin_res) if lin_res is not None else None,
                            linear_solve_time=float(lin_solve_dt) if lin_solve_dt is not None else None,
                            pc_setup_time=float(pc_setup_dt) if pc_setup_dt is not None else None,
                            pmat_build_time=float(pmat_dt) if pmat_dt is not None else None,
                            pmat_rebuilds=self._petsc_shell_pmat_rebuilds if use_pmat else None,
                            pmat_mode=self.config.petsc_shell_pmat_mode if use_pmat else None,
                            tol=self.config.tol,
                            stop_reason=("converged" if lin_conv else "linfail"),
                            nan_detected=bool(np.isnan(lin_res)) if lin_res is not None else False,
                        )
                        if len(fallback_attempts) > 1:
                            info.linear_fallbacks = fallback_attempts
                    else:
                        raise ValueError(f"Unknown linear solve method: {self.config.method}")

                except Exception as e:  # pragma: no cover - defensive
                    exception = repr(e)
                    info = SolverResult(
                        converged=False,
                        iters=0,
                        linear_iters=None,
                        linear_converged=False,
                        linear_residual=None,
                        tol=self.config.tol,
                        stop_reason=f"exception: {exception}",
                        nan_detected=False,
                    )
                    # keep u as None to signal failure; caller should treat as Optional
                    result = LinearStepResult(info=info, solve_time=0.0, u=None)
                    if step_callback is not None:
                        step_callback(result)
                    # Report outside.
                    if report_timing:
                        timer.report(sort_by="total")
                    return None, [result], timer

            with timer.section("expand_dirichlet"):
                if free is None:
                    u_full = np.asarray(u_free, dtype=float)
                else:
                    u_full = expand_dirichlet_solution(
                        u_free, free, dir_arr, dir_vals_arr, n_total
                    )

            solve_key = f"solve>{self.config.method}"
            solve_time = timer._records.get(solve_key, [0.0])[-1]
            result = LinearStepResult(
                info=info, solve_time=solve_time, u=u_full
            )
            if step_callback is not None:
                step_callback(result)

        if report_timing:
            timer.report(sort_by="total")

        return u_full, [result]
